/****** Program to grammar *********************/

:- module(grammar_and_strings,[grammar_and_strings/6,silent_predicate/1,clause_/2]).

:- use_module([misc_predicates,merlin2]).

grammar_and_strings(TheoryFile,Pos,Neg,Grammar,PosProofs,NegProofs):-
        init0,
	asserta(clause_no(0)),
	assert_silent_predicates(TheoryFile),
	see(TheoryFile),
	repeat,
	read(Clause),
	assert_rule(Clause),
	Clause == end_of_file,
	seen,
	findall(Rule,retract(rule(Rule)),Grammar), 
       	find_neg_proofs(Neg,[],NegProofs),
	find_pos_proofs(Pos,[],NegProofs,PosProofs), 
        erase_clauses.

init0 :-
    retractall(silent(_)),
    retractall(rule(_)),
    retractall(clause_(_,_)),
    retractall(clause_no(_)),
    retractall(asserted(_)),
    retractall(head_symbol(_,_)), !.

erase_clauses :-
    retract(asserted(Ref)),
    erase(Ref),
    fail.
erase_clauses.

assert_silent_predicates(File):-
	see(File),
	repeat,
	read(Clause),
	assert_silent_predicate(Clause),
	Clause == end_of_file,
	seen, !.

assert_silent_predicate(silent(Predicate)):-
	!, asserta(silent(Predicate)).
assert_silent_predicate(_).

assert_rule(end_of_file):-!.
assert_rule(silent(_)):-!.
assert_rule(:-(Head,Body)):-
	silent(Head), !,
	assertz(:-(Head,Body),Ref), 
	asserta(asserted(Ref)).	
assert_rule(:-(Head,Body)):-	
	functor(Head,Functor,Arity),
	find_right_side(Body,RHS),
	new_clause_no(No),
	assertz(rule((Functor/Arity,[No|RHS]))),
	compile_clause(:-(Head,Body),No,NewClause),
	assertz(NewClause,Ref), 
	asserta(asserted(Ref)),
	asserta(clause_(No,:-(Head,Body))),
	functor(Predicate,Functor,Arity),
	asserta((head_symbol(No,Predicate):-!)), !.
assert_rule(Head):-
	silent(Head), !,
	assertz(Head,Ref), 
	asserta(asserted(Ref)).
assert_rule(Head):-	
	functor(Head,Functor,Arity),
	new_clause_no(No),
	assertz(rule((Functor/Arity,[No]))),
	compile_clause(Head,No,NewClause),
	assertz(NewClause,Ref), 
	asserta(asserted(Ref)),
	asserta(clause_(No,Head)),
	functor(Predicate,Functor,Arity),
	asserta((head_symbol(No,Predicate):-!)), !.

find_right_side((Goal,Body),[Functor/Arity|RHS]):-
	\+ silent_predicate(Goal), !, 
	functor(Goal,Functor,Arity),
	find_right_side(Body,RHS).
find_right_side((_,Body),RHS):-
	!, find_right_side(Body,RHS).
find_right_side(Goal,[Functor/Arity]):-
	\+ silent_predicate(Goal), !, 
	functor(Goal,Functor,Arity).
find_right_side(_,[]).

silent_predicate(Goal):-
	predicate_property(Goal,built_in), !.
silent_predicate(Goal):-
	silent(Goal).

new_clause_no(NewNo):-
	retract(clause_no(OldNo)),
	NewNo is OldNo+1, 
	asserta(clause_no(NewNo)), !.

compile_clause(:-(Head,Body),N,:-(NewHead,NewBody)):-
	!, Head =.. [Predicate|Arguments1],
	app(Arguments1,[[N|Xs],Ys],Arguments2),
	NewHead =.. [Predicate|Arguments2],
	compile_body(Body,Xs,Ys,NewBody).
compile_clause(Head,N,NewHead):-
	Head =.. [Predicate|Arguments1],
	app(Arguments1,[[N|Xs],Xs],Arguments2),
	NewHead =.. [Predicate|Arguments2].

compile_body((Goal,Goals),Xs,Ys,(Goal,NewGoals)):-
	silent_predicate(Goal), !,
	compile_body(Goals,Xs,Ys,NewGoals).
compile_body((Goal,Goals),Xs,Ys,(NewGoal,NewGoals)):-
	!, Goal =.. [Predicate|Arguments1],
	app(Arguments1,[Xs,Zs],Arguments2),
	NewGoal =.. [Predicate|Arguments2],
	compile_body(Goals,Zs,Ys,NewGoals).
compile_body(Goal,Xs,Xs,Goal):-
	silent_predicate(Goal), !.
compile_body(Goal,Xs,Ys,NewGoal):-
	Goal =.. [Predicate|Arguments1],
	app(Arguments1,[Xs,Ys],Arguments2),
	NewGoal =.. [Predicate|Arguments2].

/*** Find SLD-refutations for the examples ****************/

find_neg_proofs([],_,[]):-!.
find_neg_proofs([NegEx|RestNeg],SeenProofs,NegProofs):-
	NegEx =.. [Predicate|Args],
	app(Args,[Proof,[]],NewArgs),
	NewGoal =.. [Predicate|NewArgs],
	findall(Proof, (NewGoal, \+ member(Proof,SeenProofs)), NegExProofs), 
	app(NegExProofs,RestNegProofs,NegProofs),
	app(NegExProofs,SeenProofs,NewSeenProofs),
	find_neg_proofs(RestNeg,NewSeenProofs,RestNegProofs), !.

find_pos_proofs([],PosProofs,_,NewPosProofs):-
        reverse(PosProofs,[],NewPosProofs), !.
find_pos_proofs([PosEx|RestPos],SeenPosProofs,NegProofs,PosProofs):-
	PosEx =.. [Predicate|Args],
	app(Args,[Proof,[]],NewArgs),
	NewGoal =.. [Predicate|NewArgs],
	findall(Proof, (NewGoal, \+ member(Proof,NegProofs)), PosExProofs),
	update_seen_proofs(PosEx,PosExProofs,SeenPosProofs,NewSeenPosProofs),
	find_pos_proofs(RestPos,NewSeenPosProofs,NegProofs,PosProofs), !.

update_seen_proofs(_,PosExProofs,SeenPosProofs,SeenPosProofs):-
	member(PosProof,PosExProofs),
	member(PosProof,SeenPosProofs), !.
update_seen_proofs(_,[PosExProof|_],SeenPosProofs,[PosExProof|SeenPosProofs]):-!.
update_seen_proofs(PosEx,_,SeenPosProofs,SeenPosProofs):-
        write_int('*** Warning! The following positive example has been excluded:\n'),
        write_clauses([PosEx]).
