/********** Grammar intersection ****************/

:- module(intersection,[grammar_intersection/5]).

:- use_module(misc_predicates).
    
grammar_intersection(StartSymbol,StartState,CFG,Automaton,NewCFG):-
        init,
	assert_states(Automaton),
	produce_symbol_table(CFG,Automaton),
	grammar_intersection2([(StartSymbol,StartState,empty)],
			      [(StartSymbol,StartState,empty)],CFG,Automaton,
			      TempCFG1), 
	remove_dead_rules(TempCFG1,[],TempCFG2),
	new_rules(TempCFG1,TempCFG2,NewCFG), !.

init :-
    retractall(poss_int(_,_)),
    retractall(state(_)), !.

assert_states(Automaton):-
	member((State,_,_),Automaton),
	\+ state(State),
	assert(state(State)),
	fail.
assert_states(Automaton):-
	member((_,_,State),Automaton),
	\+ state(State),
	assert(state(State)),
	fail.
assert_states(_).

produce_symbol_table(CFG,Automaton):-
	member((Left,[Sigma|_]),CFG),
	assert_poss_int(Left,Sigma,Automaton),	
	fail.
produce_symbol_table(_,_).

assert_poss_int(Left,Sigma,Automaton):-
        member((State,Sigma,_),Automaton),
	\+ poss_int(Left,State),
	asserta(poss_int(Left,State)).

grammar_intersection2([],_,_,_,[]):-!.
grammar_intersection2([Symbol|Symbols],SeenSymbols,CFG,Automaton,NewCFG):-
	find_rule_intersection(Symbol,CFG,Automaton,FirstRules),
	new_symbols(FirstRules,SeenSymbols,NewSymbols),
	app(FirstRules,RestRules,NewCFG),
	app(Symbols,NewSymbols,RemSymbols),
	app(SeenSymbols,NewSymbols,NewSeenSymbols),
	grammar_intersection2(RemSymbols,NewSeenSymbols,CFG,
			      Automaton,RestRules), !.

find_rule_intersection(_,[],_,[]):- !.
find_rule_intersection((Left,State1,Rem),[(Left,Right)|CFG],
			Automaton,NewRules):-
	rule_intersection(State1,Automaton,Left,Right,Rem,FirstRules), !,
	app(FirstRules,RestRules,NewRules),
	find_rule_intersection((Left,State1,Rem),CFG,Automaton,RestRules).
find_rule_intersection((Left,State1,Rem),[_|CFG],Automaton,NewRules):-
	find_rule_intersection((Left,State1,Rem),CFG,Automaton,NewRules).

rule_intersection(_,[],_,_,_,[]):-!.
rule_intersection(State1,[(State1,Sigma,final(_))|Automaton],Left,[Sigma],
		  empty,[((Left,State1,empty),[Sigma])|RestRules]):-
    !, rule_intersection(State1,Automaton,Left,[Sigma],empty,RestRules).
rule_intersection(State1,[(State1,Sigma,State2)|Automaton],Left,[Sigma],State2,
		  [((Left,State1,State2),[Sigma])|RestRules]):-
    !, rule_intersection(State1,Automaton,Left,[Sigma],State2,RestRules).
rule_intersection(State1,[(State1,Sigma,State2)|Automaton],Left,[Sigma|Right],Rem,
		  NewRules):-
    Right \== [], !,
    findall(((Left,State1,Rem),[Sigma|NewRight]), 
	new_rule(Right,State2,Rem,NewRight), 
	FirstRules),
    app(FirstRules,RestRules,NewRules),
    rule_intersection(State1,Automaton,Left,[Sigma|Right],Rem,RestRules).
rule_intersection(State1,[_|Automaton],Left,Right,Rem,NewRules):-
	rule_intersection(State1,Automaton,Left,Right,Rem,NewRules).

new_rule([Right],State2,Remainder,[(Right,State2,Remainder)]):-
	poss_int(Right,State2), !.
new_rule([Right|Rest],State2,Remainder,[(Right,State2,State3)|NewRest]):-
	poss_int(Right,State2),
	state(State3),
	new_rule(Rest,State3,Remainder,NewRest).

new_symbols([],_,[]):-!.
new_symbols([(_,[_|RHS])|SelectedRules],SeenNTs,SelectedNTs):-
	new_nts(RHS,SeenNTs,NewNTs),
	app(NewNTs,RestNTs,SelectedNTs),
	app(SeenNTs,NewNTs,NewSeenNTs),
	new_symbols(SelectedRules,NewSeenNTs,RestNTs), !.

new_nts([],_,[]):-!.
new_nts([NT|RHS],SeenNTs,[NT|NewNTs]):-
	\+ member(NT,SeenNTs), !,
	new_nts(RHS,[NT|SeenNTs],NewNTs).
new_nts([_|RHS],SeenNTs,NewNTs):-
	new_nts(RHS,SeenNTs,NewNTs).

remove_dead_rules(OrigCFG,TempCFG,NewCFG):-
	select((LHS,[Sigma|RHS]),OrigCFG,RestCFG),
	\+ (member(Symbol,RHS),	\+ member((Symbol,_),TempCFG)), !,
	remove_dead_rules(RestCFG,[(LHS,[Sigma|RHS])|TempCFG],NewCFG).
remove_dead_rules(_,NewCFG,NewCFG).

new_rules([],_,[]):-!.
new_rules([Rule|Rules],NotDeadRules,[Rule|NewRules]):-
	member(Rule,NotDeadRules), !,
	new_rules(Rules,NotDeadRules,NewRules).
new_rules([_|Rules],NotDeadRules,NewRules):-
	new_rules(Rules,NotDeadRules,NewRules).
