% Written by Henrik Bostrm in SICStus Prolog 3 #5

:- module(merlin2,[init/0,run_merlin1/0,run_merlin2/0,write_int/1,interpreter/1,
		   write_clauses/1,theory/1,examples/1,batch/1,factor/1,look/1,
		   det/1, help_text/0]).

:- use_module([library(tcltk),merlin1_module,merlin2_module,det_merlin2_module]).


start :- 
	tk_new([name('MERLIN 2.0')],I),
	assert(interpreter(I)),
	tcl_eval(I,'source merlin2.tcl',_),
	tk_main_loop,
	tcl_delete(I),
	retractall(interpreter(_)),!.

/*********** Initiation ***************/

init :-
    retractall(theory_file(_)),
    retractall(example_file(_)),
    retractall(batch_size(_)),
    retractall(lambda(_)),
    retractall(deterministic(_)), !.
    
theory(File):-
	assert(theory_file(File)),!.
examples(File):-
	assert(example_file(File)),!.
batch(BatchSize):-
        name(BatchSize,L),
        name(N,L),
        assert(batch_size(N)), !.
factor(Lambda):-
        name(Lambda,L),
        name(N,L),
        assert(lambda(N)), !.
look(LookAhead):-
        name(LookAhead,L),
        name(N,L),
        assert(look_ahead(N)), !.
det(Det):-
    name(Det,L),
    name(N,L),
    assert(deterministic(N)), !.
    

/**************** Main goal*******************/

run_merlin1 :-
	theory_file(TheoryFile),
	example_file(ExampleFile),
	write_int('\n==================== MERLIN 2.0 Session ===================='), 
        write_int('\n'),
	write_int('Theory file: '), write_int(TheoryFile),write_int('\n'),
	write_int('Example file: '), write_int(ExampleFile),write_int('\n'),
        write_int('Minimising the size of the induced automaton.'), write_int('\n'),
	statistics(runtime,[OldTime,_]),
%	write_int('Reading examples.'), write_int('\n'),
	read_examples(ExampleFile,Pos,Neg),
%	statistics(runtime,[TempTime,_]),
%	ReadTime is (TempTime-OldTime)/1000,
%	write_int('\tTime taken:'), write_int(ReadTime), write_int(' s.'), write_int('\n'),
	target_predicate_symbol(ExampleFile,TargetPredicate),
        merlin1_module:start(TheoryFile,TargetPredicate,Pos,Neg,NewClauses),
	statistics(runtime,[NewTime,_]),
	TotalTime is (NewTime-OldTime)/1000,
	write_int('Total time taken: '), write_int(TotalTime), write_int(' s.'),
	write_int('\n'),
         write_int('- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n'),
	write_int('Hypothesis:'),write_int('\n'),
	write_clauses(NewClauses),
	write_int('\n').
run_merlin1 :-
        write_int('- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n'),
	write_int('\nNo hypothesis produced!\n\n'),
	seen.

run_merlin2 :-
	theory_file(TheoryFile),
	example_file(ExampleFile),
        batch_size(BatchSize),
	lambda(Lambda),
        look_ahead(LookAhead),
	write_int('==================== MERLIN 2.0 Session ===================='), 
        write_int('\n'),
	write_int('Theory file: '), write_int(TheoryFile),write_int('\n'),
	write_int('Example file: '), write_int(ExampleFile),write_int('\n'),
        write_int('Maximising the posterior probability of the induced automaton.'), 
        write_int('\n'),
	write_int('Batch size: '), write_int(BatchSize),write_int('\n'),
	write_int('Lambda factor: '), write_int(Lambda),
        write_int('\n'),
	write_int('Look ahead: '), write_int(LookAhead),write_int('\n'),
	statistics(runtime,[OldTime,_]),
%	write_int('Reading examples.'), write_int('\n'),
	read_examples(ExampleFile,Pos,Neg),
%	statistics(runtime,[TempTime,_]),
%	ReadTime is (TempTime-OldTime)/1000,
%	write_int('\tTime taken:'), write_int(ReadTime), write_int(' s.'), write_int('\n'),
	target_predicate_symbol(ExampleFile,TargetPredicate),
        start2(TheoryFile,TargetPredicate,Pos,Neg,BatchSize,Lambda,LookAhead,NewClauses),
	statistics(runtime,[NewTime,_]),
	TotalTime is (NewTime-OldTime)/1000,
	write_int('Total time taken: '), write_int(TotalTime), write_int(' s.'),
	write_int('\n'),
         write_int('- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n'),
	write_int('Hypothesis:'),write_int('\n'),
	write_clauses(NewClauses),
	write_int('\n').
run_merlin2 :-
         write_int('- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n'),
	write_int('\nNo hypothesis produced!\n\n'),
	seen,
	erase_clauses.

start2(TheoryFile,TargetPredicate,Pos,Neg,BatchSize,Lambda,LookAhead,NewClauses):-
        deterministic(1), !,
        write_int('Inducing deterministic automaton.\n'),
	det_merlin2_module:start(TheoryFile,TargetPredicate,Pos,Neg,BatchSize,Lambda,
				 LookAhead,NewClauses).
start2(TheoryFile,TargetPredicate,Pos,Neg,BatchSize,Lambda,LookAhead,NewClauses):-
        write_int('Inducing non-deterministic automaton.\n'),
	merlin2_module:start(TheoryFile,TargetPredicate,Pos,Neg,BatchSize,Lambda,
				 LookAhead,NewClauses).


/**************************************************/

read_examples(ExampleFile,Pos,Neg):-
	see(ExampleFile),
	read_examples(Pos,Neg),
	seen, !.
	
read_examples(NewPos,NewNeg):-
	read(Example),
	Example \== end_of_file, !,
	update_pos_and_neg(Example,NewPos,NewNeg,RestPos,RestNeg),
	read_examples(RestPos,RestNeg).
read_examples([],[]).

update_pos_and_neg((:- Example),Pos,[Example|Neg],Pos,Neg):-!.
update_pos_and_neg(Example,[Example|Pos],Neg,Pos,Neg).

target_predicate_symbol(ExampleFile,PredicateSymbol):-
	see(ExampleFile),
	read(Example),
	seen,
        get_predicate_symbol(Example,PredicateSymbol), !.

get_predicate_symbol((:- Example),Functor/Arity):-
        !, functor(Example,Functor,Arity).
get_predicate_symbol(Example,Functor/Arity):-
        !, functor(Example,Functor,Arity).

/********** I/O *********************************/

write_clauses([]).
write_clauses([Clause|Clauses]):-
	clause_to_digits(Clause,Digits),
	name(String,Digits),
	write_int(String),
	write_clauses(Clauses).

clause_to_digits(:-(Head,Body),Digits):-
	!, goal_to_digits(Head,65,Next,Digits1),
	body_to_digits(Body,Next,Digits2),
	app(Digits1,[32,58,45,10|Digits2],Digits).
clause_to_digits(Head,Digits):-
	!, goal_to_digits(Head,65,_,Digits1),
	app(Digits1,[46,10],Digits).

goal_to_digits(Goal,Next1,Next2,Digits):-
	instantiate_vars(Goal,Next1,Next2),
	term_to_digits(Goal,Digits).

instantiate_vars(Term,Next,Next):-
	atomic(Term), !.
instantiate_vars(Var,Next1,Next2):-
	var(Var), !,
	bind_var(Next1,Var),
	Next2 is Next1+1.
instantiate_vars(Term,Next1,Next2):-
	functor(Term,_,A),
	instantiate_vars2(Term,1,A,Next1,Next2).

bind_var(Next,Var):-
	Next < 91, !,
	name(Var,[Next]).
bind_var(Next,Var):-
	No is Next-90,
	name(No,L),
	name(Var,[65|L]).

instantiate_vars2(Term,A,A,Next1,Next2):-
	!, arg(A,Term,Arg),
	instantiate_vars(Arg,Next1,Next2).
instantiate_vars2(Term,A1,A2,Next1,Next2):-
	arg(A1,Term,Arg),
	instantiate_vars(Arg,Next1,Next3),
	A3 is A1+1,
	instantiate_vars2(Term,A3,A2,Next3,Next2).

term_to_digits(Term,Digits):-
	atomic(Term), 
	name(Term,Digits),!.
term_to_digits([Head|Tail],Digits1):-
	!, term_to_digits(Head,Digits2),
	list_to_digits(Tail,Digits3),
	app([91|Digits2],Digits3,Digits1).
term_to_digits((Head,Tail),Digits1):-
	!, term_to_digits(Head,Digits2),
	pair_to_digits(Tail,Digits3),
	app([40|Digits2],Digits3,Digits1).
term_to_digits(Term,Digits):-
	functor(Term,F,A),
	name(F,Digits1),
	term_to_digits2(Term,1,A,Digits2),
	app(Digits1,[40|Digits2],Digits).

list_to_digits([],[93]):-!.
list_to_digits([Head|Tail],Digits1):-
	!, term_to_digits(Head,Digits2),
	list_to_digits(Tail,Digits3),
	app([44|Digits2],Digits3,Digits1).
list_to_digits(Tail,Digits1):-
	!, term_to_digits(Tail,Digits2),
	app([124|Digits2],[93],Digits1).

pair_to_digits((Head,Tail),Digits1):-
	!, term_to_digits(Head,Digits2),
	pair_to_digits(Tail,Digits3),
	app([44|Digits2],Digits3,Digits1).
pair_to_digits(Tail,Digits1):-
	!, term_to_digits(Tail,Digits2),
	app([44|Digits2],[41],Digits1).

term_to_digits2(Term,A,A,Digits):-
	!, arg(A,Term,Arg),
	term_to_digits(Arg,Digits1),
	app(Digits1,[41],Digits).
term_to_digits2(Term,A1,A2,Digits):-
	arg(A1,Term,Arg),
	term_to_digits(Arg,Digits1),
	A3 is A1+1,
	term_to_digits2(Term,A3,A2,Digits2),
	app(Digits1,[44|Digits2],Digits).

body_to_digits((Goal,Body),Next1,Digits):-
	!, goal_to_digits(Goal,Next1,Next2,Digits1),
	body_to_digits(Body,Next2,Digits2),
	app([9|Digits1],[44,10|Digits2],Digits).
body_to_digits(Goal,Next,Digits):-
	!, goal_to_digits(Goal,Next,_,Digits1),
	app([9|Digits1],[46,10],Digits).

app([],L,L):- !.
app([X|L1],L2,[X|L3]):-
	app(L1,L2,L3).

write_int(String):-
	interpreter(I),
	tcl_eval(I,['.t',insert,end,br([String])],_),
	tk_do_one_event, !.

/*****************************************************/

help_text :-
	interpreter(I),
	tcl_eval(I,'toplevel .help',_),
	tcl_eval(I,'text .help.text -width 75 -yscrollcommand ".help.scroll set"',_),
	tcl_eval(I,'scrollbar .help.scroll -command ".help.text yview"',_),
	tcl_eval(I,'pack .help.scroll -side right -fill y',_),
	tcl_eval(I,'pack .help.text -side left',_),
	help_text(String),
	tcl_eval(I,'raise .help',_),
%	tk_do_one_event,
	tcl_eval(I,['.help.text',insert,end,br([String])],_),!.
%	tk_do_one_event, !.

help_text(
'MERLIN 2.0 HELP
=================================================================
In order to run MERLIN 2.0, two input files are needed:

Theory file - a Prolog program in Edinburgh syntax that gives an
              overly general definition of a target predicate

Example file - positive and (optionally) negative instances of the
               target predicate in the form of unit clauses. The
               negative instances are prefixed by ":-".

MERLIN 2.0 first tries to find SLD-refutations for the positive
examples using the program in the Theory file.  Wieving the sequences
of input clauses in these refutations as strings in a formal language,
MERLIN 2.0 induces a finite-state automaton that can generate all
positive sequences (and none of the negative sequences). The automaton
is then used to specialise the overly general hypothesis, a process in
which new predicates are invented.

There are two modes in which MERLIN 2.0 may be run:

1. Learning from positive and negative examples
-----------------------------------------------

This mode employs a hill-climbing technique for finding the minimal
automaton that can generate all positive and no negative sequences.
Note that running MERLIN 2.0 under this mode results in the same
hypotheses as produced by MERLIN 1.0. This mode requires no further
parameters to be specified, but is not very useful in the absence of
negative examples.

2. Learning from positive examples only
---------------------------------------

This mode employs an incremental hill-climbing technique for inducing
the most probable Hidden Markov Model structure that can generate the
positive sequences X, by maximising log P(Ms) + log P(X|Ms), where log
P(Ms) is the natural logarithm of the prior of the model structure Ms
and log P(X|Ms) is the logarithm of the probability of the sequences
given Ms. Note that in the current version of the system, the negative
examples are completely disregarded in this mode, e.g. it is not
checked whether negative examples are covered or not.

In this mode, the following options must be set:

Deterministic HMM     - results in that the induced HMM is kept 
                        deterministic (the default is ON)

Batch size            - specifies the number of examples in 
                        each batch (default is 1)

Lambda                - is used to prevent overgeneralisation
                        when having considered too few examples,
                        where the logarithm of the prior is multiplied 
                        with Lambda (default is 1)

Look ahead            - specifies the number of steps the 
                        Hill-climbing search should continue
                        after having found a (local) maximum
                        (default is 1)

For both modes, the resulting hypothesis is presented on the left side
of the main window. In the first mode, MERLIN 2.0 assumes that each
positive example has at least one SLD-refutation such that there is no
negative example with an SLD-refutation that has the same sequence of
input clauses (exceptions to this are reported).

MERLIN 2.0 was implemented in SICStus Prolog 3 #5 by Henrik Bostrm.

').

/*****************************************************/
:- start.

