/********** Miscellaneous predicates ****************/

:- module(misc_predicates,[app/3,append/3,common_element/2,delete/3,
			   delete_chk/3,member/2,member_chk/2, reverse/3,
			   select/3,select_rest/3,sum/3,union/3]).
app([],L,L):- !.
app([X|L1],L2,[X|L3]):-
	app(L1,L2,L3).

append([],L,L).
append([X|L1],L2,[X|L3]):-
	append(L1,L2,L3).

common_element(L1,L2):-
	member(X,L1),
	member(X,L2), !.

delete(_,[],[]):-!.
delete(X,[X|L1],L2):-
	!, delete(X,L1,L2).
delete(X,[Y|L1],[Y|L2]):-
	delete(X,L1,L2).

delete_chk(X,[X|L],L):-!.
delete_chk(X,[Y|L1],[Y|L2]):-
	delete_chk(X,L1,L2).

member(X,[X|_]).
member(X,[_|L]):- 
	member(X,L).

member_chk(X,[X|_]):-!.
member_chk(X,[_|L]):- 
	member_chk(X,L).

reverse([],L,L):-!.
reverse([X|L1],L2,L3):-
	reverse(L1,[X|L2],L3).

select(X,[X|L],L).
select(X,[Y|L1],[Y|L2]):-
	select(X,L1,L2).
	
select_rest(X,[X|L],L).
select_rest(X,[_|L1],L2):-
	select_rest(X,L1,L2).

sum([],S,S):-!.
sum([N|Ns],S1,S2):-
    S3 is S1+N,
    sum(Ns,S3,S2).

union([],L,L):-!.
union([X|L1],L2,L3):-
	member(X,L2),
	union(L1,L2,L3), !.
union([X|L1],L2,[X|L3]):-
	union(L1,L2,L3), !.
